
 /*   ORXONOX - the hottest 3D action shooter ever to exist
 *                    > www.orxonox.net <
 *
 *
 *   License notice:
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *   Author:
 *      Fabian 'x3n' Landau
 *   Co-authors:
 *      Simon Miescher
 *
 */

/*

*
*
* An asteroid which can be destroyed. Some smaller asteroids are created and a pickup 
* spawns. 
*
*
*

*/

/*
Veraenderungstagebuch
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

KNACKPUNKTE:
o Collision Shape-Problem (Funktioniert nur, wenn im Konstruktor)
o Maximale Hitpunkte = 200?

o Neue Asteroiden generieren -> Absturz
---> Weil Elterndings zerstoert wird? 
---> Nicht moeglich, zur Laufzeit zu generieren?

o Pickups: setMaxSpawned funktioniert nicht


OFFEN: 
o Add custom pickup 'resources'. Weird template stuff -> Problems setting 'size' and other properties? setNumber of Resources. 
o Explosion parts
o custom HUD

HANDBUCH:
o im Level-File includes/pickups.oxi importieren. 

Anpassungen Pickup-Zeug: 
o Pickup.h: Zugriffsänderung createSpawner
o PickupSpawner.h: Zugriffsrechte setPickupTemplateName() und setMaxSpawnedItems()

ERLEGT:
o Rand() geht bis zu riesigen Nummern! 
o Pickupgenerierung: vgl. Anpassungen. 
o Dummes fisch-- statt fisch++ , Endlosschleife! 
o Dynamische Definition -> putStuff-Methode, Werte noch nicht durchgesickert. 
o Error-Nachricht: Einfach Argumente leer lassen. 
*/


#include "../../orxonox/worldentities/pawns/Pawn.h"
#include "../../orxonox/worldentities/WorldEntity.h"

#include "AsteroidMinable.h"

#include <algorithm>

#include "core/CoreIncludes.h"
#include "core/GameMode.h"
#include "core/XMLPort.h"
#include "core/EventIncludes.h"
#include "network/NetworkFunction.h"

// #include "infos/PlayerInfo.h"
// #include "controllers/Controller.h"
// #include "gametypes/Gametype.h"
// #include "graphics/ParticleSpawner.h"
// #include "worldentities/ExplosionChunk.h"
// #include "worldentities/ExplosionPart.h"

// #include "core/object/ObjectListIterator.h"
// #include "controllers/FormationController.h"

#include "../pickup/items/HealthPickup.h"
#include "../pickup/PickupSpawner.h"
#include "../pickup/Pickup.h"
//#include "../pickup/pickup ..... pickupable
#include "../objects/collisionshapes/SphereCollisionShape.h"
#include "../../orxonox/graphics/Model.h"






namespace orxonox
{
    RegisterClass(AsteroidMinable);

    AsteroidMinable::AsteroidMinable(Context* context) : Pawn(context)
    {
        RegisterObject(AsteroidMinable);

        // Old stuff from pawn
        this->setRadarObjectColour(ColourValue(1.0f, 1.0f, 0.0f, 1.0f));
        this->setRadarObjectShape(RadarViewable::Shape::Dot);
        this->setCollisionType(WorldEntity::CollisionType::Dynamic);
        this->bAlive_ = true;
        this->bVulnerable_ = true;

        // Default Values
        this->generateSmaller = true;
        this->size = 10; // customSize
        this->context = context;
        this->initialised = false;
        //this->roll = rand()*5; //etwas Drehung. richtige Variable




        // DELETE if other stuff works! (wrong size etc.)
        SphereCollisionShape* cs = new SphereCollisionShape(this->context);
        cs->setRadius((this->size)*2); //OFFEN: Feinabstimmung der Radien 
        this->attachCollisionShape(cs); 




        // Old from Pawn
        this->registerVariables();

    }

    AsteroidMinable::~AsteroidMinable()
    {

    }

    void AsteroidMinable::putStuff(){

        // Add Model
        //<Model position="0,-40,40" yaw="90" pitch="-90" roll="0" scale="4" mesh="ast6.mesh" />
        Model* hull = new Model(this->context);
        // random one of the 6 shapes
        char meshThingy[] = "";
        sprintf(meshThingy, "ast%.0f.mesh", (round((5*(double)rand()/(double)RAND_MAX))+1)); //    sprintf(str, "Value of Pi = %f", M_PI);
        hull->setMeshSource(meshThingy);
        hull->setScale(this->size);
        this->attach(hull);

        // Collision shape
        SphereCollisionShape* cs = new SphereCollisionShape(this->context);
        cs->setRadius((this->size)*2); //OFFEN: Feinabstimmung der Radien. 2.5 in AsteroidField.lua
        this->attachCollisionShape(cs); 


        this->initialised=true;
    }

    void AsteroidMinable::XMLPort(Element& xmlelement, XMLPort::Mode mode)
    {
        SUPER(AsteroidMinable, XMLPort, xmlelement, mode);
        //        XMLPortParam(PickupSpawner, "pickup", setPickupTemplateName, getPickupTemplateName, xmlelement, mode);
        XMLPortParam(AsteroidMinable, "size", setSize, getSize, xmlelement, mode);

    }

    void AsteroidMinable::XMLEventPort(Element& xmlelement, XMLPort::Mode mode)
    {
        SUPER(AsteroidMinable, XMLEventPort, xmlelement, mode);

        XMLPortEventState(AsteroidMinable, BaseObject, "vulnerability", setVulnerable, xmlelement, mode);
    }

    void AsteroidMinable::registerVariables()
    {
        registerVariable(this->bAlive_,            VariableDirection::ToClient);
        registerVariable(this->bVulnerable_,       VariableDirection::ToClient);
        registerVariable(this->health_,            VariableDirection::ToClient);
        registerVariable(this->maxHealth_,         VariableDirection::ToClient);

        registerVariable(this->size, VariableDirection::ToClient);
        registerVariable(this->generateSmaller, VariableDirection::ToClient);

        registerVariable(this->initialised, VariableDirection::ToClient);
        //registerVariable(this->context, VariableDirection::ToClient); // can't link that since it's a context


            //         float size;
            // bool generateSmaller;
            // bool initialised;

            // Context* context;
    }

    void AsteroidMinable::tick(float dt)
    {
        // SUPER(Pawn, tick, dt);
        if(!(this->initialised)){this->putStuff();}

        if(this->health_ <=0){this->death();}

    }

    void AsteroidMinable::setSize(float s){
        this->size = s;
        this->health_ = 5*s;// capped at 200 in pawn or smth?
        this->setHealth(health_);
    }

    float AsteroidMinable::getSize(){
        return this->size;
    }

    void AsteroidMinable::death() //ueberschreiben
    {

        // Stuff that can be harvested. 
        PickupSpawner* thingy = new PickupSpawner(this->context);

        // OFFEN: more precise size relation in custom resource pickup. 
        char tname[] = ""; // can-t overwrite strings easily in C (strcat etc.)
        if(this->size <= 5){
            strcat(tname, "smallmunitionpickup");
        }else if(this->size <= 20){
            strcat(tname, "mediummunitionpickup");
        }else{
            strcat(tname, "hugemunitionpickup");
        }
        //orxout() << "Zeugsname:" << tname << endl;
        thingy->setPickupTemplateName(tname);
        thingy->setPosition(this->getPosition());
        //thingy->setMaxSpawnedItems(1); // Somehow doesn-t work yet. 
        thingy->setRespawnTime(0.1f);// instantly
        
        orxout() << "AsteroidMining::Death(): Passed Pickup stuff!" << endl;





        // Smaller Parts = 'Children'
        if(this->generateSmaller){
            this->spawnChildren();
        }





        // OFFEN: Sauber kapputten
        // just copied other stuff:
        this->setHealth(1);
        this->bAlive_ = false;
        this->destroyLater();
        this->setDestroyWhenPlayerLeft(false);


        // pawn -> addExplosionPart
        // this->goWithStyle();
        orxout() << "Wieder retour in death() geschafft. " << endl;

        // if (this->getGametype() && this->getGametype()->allowPawnDeath(this, this->lastHitOriginator_))
        // {
        //     // Set bAlive_ to false and wait for destroyLater() to do the destruction
        //     this->bAlive_ = false;
        //     this->destroyLater();

        //     this->setDestroyWhenPlayerLeft(false);

        // }
    }


void AsteroidMinable::spawnChildren(){
    
    // Spawn smaller Children 
    int massRem = this->size-1; //some mass is lost
    int num = round((massRem-1)*(double)rand() / (double)RAND_MAX)+1; // random number of children, at least one // Tweak towards bigger junks?
    num = 1; // zum Testen mal nur einen generieren. 
    
    massRem = massRem-num;    
    int extra = 0;


    orxout() << "Number of Children: " << num << endl;


    for(int fisch=num; fisch>=1; fisch--){
        // to distribute remaining mass
        if(fisch==1){ 
            extra = massRem;
        }else{
            extra = round(massRem*(double)rand() / (double)RAND_MAX);
            massRem = massRem-extra;

        }

        orxout() << "Mass chosen: " << extra+1 << endl;

        orxout() << "AsteroidMining::spawnChildren(): Inside for-loop." << endl;

        //Spawn this child  Game crashes!
        AsteroidMinable* child = new AsteroidMinable(this->context);
        // if(!(child == nullptr)){//Errorgebastel

        // Position zu spaet gesetzt? Tick nicht atomar -> falsche Groesse evtl?

            if(child == nullptr){
                orxout(internal_error, context::pickups) << "Weird, can't create new AsteroidMinable." << endl;
            }

            child->setSize(extra + 1);
                
            //OFFEN:Kollision der Kinder verhindern
            //Relativ zu Elternteil automatisch?
            //Typ position:rand()*Vektoriwas?
            //pawn->getWorldPosition() + Vector3(30,0,-30);
            child->setPosition(this->getPosition() + Vector3(num*5, 0, 0));


        orxout() << "Done: Creating new Asteroid" << endl;

        // }


                //         Pawn* pawn = this->carrierToPawnHelper();
                // if(pawn == nullptr) // If the PickupCarrier is no Pawn, then this pickup is useless and therefore is destroyed.
                //     this->Pickupable::destroy();

                // //Attach to pawn
                // Drone* drone = new Drone(pawn->getContext()); // this is neccessary because the projectiles fired need a valid creator for the particlespawner (when colliding against something)
                // drone->addTemplate(this->getDroneTemplate());


    }
    orxout() << "Leaving spawnChildren() method. " << endl;
}




}





/*
    void DronePickup::changedUsed(void)
    {
        SUPER(DronePickup, changedUsed);

        // If the pickup has transited to used.
        if(this->isUsed())
        {

                Pawn* pawn = this->carrierToPawnHelper();
                if(pawn == nullptr) // If the PickupCarrier is no Pawn, then this pickup is useless and therefore is destroyed.
                    this->Pickupable::destroy();

                //Attach to pawn
                Drone* drone = new Drone(pawn->getContext()); // this is neccessary because the projectiles fired need a valid creator for the particlespawner (when colliding against something)
                drone->addTemplate(this->getDroneTemplate());

                Controller* controller = drone->getController();
                DroneController* droneController = orxonox_cast<DroneController*>(controller);
                if(droneController != nullptr)
                {
                    droneController->setOwner(pawn);
                }

                Vector3 spawnPosition = pawn->getWorldPosition() + Vector3(30,0,-30);
                drone->setPosition(spawnPosition);

                // The pickup has been used up.
                this->setUsed(false);
        }
        else
        {
            // If either the pickup can only be used once or it is continuous and used up, it is destroyed upon setting it to unused.
            if(this->isOnce() || (this->isContinuous() ))
            {
                this->Pickupable::destroy();
            }
        }
    }
*/